@@ -1,5 +1,5 @@
 package Devel::Camelcadedb;
-our $VERSION = "1.6.1.2";
+our $VERSION = "1.6.1.3";
 
 # http://perldoc.perl.org/DB.html
 # http://perldoc.perl.org/perldebug.html
@@ -15,14 +15,16 @@ use PadWalker qw/peek_my peek_our/;
 use Scalar::Util;
 use Encode;
 use overload;
+use Hash::StoredIterator;
 #use Carp;
 
 #sub FLAG_REPORT_GOTO() {0x80;}
 
-sub STEP_CONTINUE() {0;}
-sub STEP_INTO() {1;}
-sub STEP_OVER() {2;}
-
+use constant {
+    STEP_CONTINUE => 0,
+    STEP_INTO     => 1,
+    STEP_OVER     => 2,
+};
 
 # Each array @{"::_<$filename"} holds the lines of $filename for a file compiled by Perl. The same is also true for evaled
 # strings that contain subroutines, or which are currently being executed. The $filename for evaled strings looks like
@@ -118,6 +120,12 @@ my $_script_charset = 'utf8';   # all sources and strings without utf flag will
 
 my $_skip_run_stop = 0; # flag for skipping forced stop on run phase
 
+# this enables pausing and breakpoints setting while script is running, gives moderate overhead
+my $_enable_noninteractive_mode = 0;
+
+# this enables attemp to set a breakpoint on each leaving/entering sub, gives large overhead, proportional number of breakpoints
+my $_enable_compile_time_breakpoints = 0;
+
 my $_debug_socket;
 my $_debug_packed_address;
 my IO::Select $_debug_socket_select;
@@ -409,7 +417,10 @@ sub _get_reference_subelements
         }
         elsif ($reftype eq 'HASH')
         {
+            my $hash_iterator = Hash::StoredIterator::hash_get_iterator( $source_data );
             my @keys = sort keys %$source_data;
+            Hash::StoredIterator::hash_set_iterator( $source_data, $hash_iterator );
+
             if ($#keys >= $offset)
             {
                 my $last_index = $offset + $size;
@@ -552,7 +563,10 @@ sub _get_reference_descriptor
     }
     elsif ($reftype eq 'HASH')
     {
+        my $hash_iterator = Hash::StoredIterator::hash_get_iterator( $value );
         $size = scalar keys %$value;
+        Hash::StoredIterator::hash_set_iterator( $value, $hash_iterator );
+
         $value = sprintf "size = %s", $size;
         $expandable = $size ? \1 : \0;
     }
@@ -624,57 +638,6 @@ sub _escape_scalar
     return $scalar;
 }
 
-sub _render_variables
-{
-    my ($vars_hash) = @_;
-    my $result = '';
-
-    foreach my $variable (keys %$vars_hash)
-    {
-        my $value = $vars_hash->{$variable};
-
-        my $reftype = Scalar::Util::reftype $value;
-        my $ref = ref $value;
-        my $appendix = '';
-
-        if ($reftype eq 'SCALAR')
-        {
-            $value = $$value // '_UNDEF_';
-        }
-        elsif ($reftype eq 'ARRAY')
-        {
-            $appendix = sprintf '[%s]', scalar @$value;
-            my @elements = @$value;
-            if (@elements > 1000)
-            {
-                splice @elements, 3; # should be a 1000 here and method to get the rest
-                push @elements, '...'
-            }
-            $value = sprintf "(%s)", join ',', map { $_ // '__UNDEF__'} @elements;
-        }
-        elsif ($reftype eq 'HASH')
-        {
-            my @elements = sort keys %$value; # sorting is necessary to iterate and get data frame
-            $appendix = sprintf '{%s}', scalar @elements;
-
-            if (@elements > 1000)
-            {
-                splice @elements, 3; # should be a 1000 here and method to get the rest
-                push @elements, '...'
-            }
-            $value = sprintf "(\n%s\n)", join ",\n", map { "\t\t".$_." => ".$value->{$_}} @elements;
-        }
-        elsif (ref $value eq 'REF')
-        {
-            $value = $$value;
-        }
-
-        $result .= "\t$ref $variable$appendix = $value\n";
-    }
-
-    return $result;
-}
-
 sub _get_current_stack_frame
 {
     return $_stack_frames->[-1];
@@ -975,7 +938,7 @@ sub _enter_frame
         single  => $old_db_single,
     };
     push @{$_stack_frames}, $new_stack_frame;
-    _apply_queued_breakpoints() if $ready_to_go;
+    _set_break_points_for_files() if $_enable_compile_time_breakpoints && $ready_to_go;
     return $new_stack_frame;
 }
 
@@ -987,7 +950,6 @@ sub _exit_frame
     $frame_prefix = $frame_prefix_step x (scalar @$_stack_frames);
     _report "Leaving frame %s, setting single to %s", (scalar @$_stack_frames + 1),
         $frame->{single} if $_debug_sub_handler && $_dev_mode;
-    _apply_queued_breakpoints() if $ready_to_go;
     $DB::single = $frame->{single};
     $_internal_process = 0;
 }
@@ -1015,40 +977,6 @@ sub _get_perl_file_id_by_real_path
 }
 
 
-sub _get_perl_line_breakpoints_map_by_file_id
-{
-    my ($file_id) = @_;
-
-    unless (defined $file_id)
-    {
-        _dump_stack && _dump_frames;
-        die "Unitialized file id";
-    }
-
-    my $glob = $::{"_<$file_id"};
-    return $glob ? *$glob{HASH} : undef;
-}
-
-sub _get_perl_line_breakpoints_map_by_real_path
-{
-    my ($real_path) = @_;
-    my $perl_file_id = _get_perl_file_id_by_real_path( $real_path );
-    return $perl_file_id ? _get_perl_line_breakpoints_map_by_file_id( $perl_file_id ) : undef;
-}
-
-sub _get_perl_source_lines_by_file_id
-{
-    my ($file_id) = @_;
-    return unless $file_id;
-    my $glob = $::{"_<$file_id"};
-    return $glob && *$glob{ARRAY} && scalar @{*$glob{ARRAY}} ? *$glob{ARRAY} : undef;
-}
-
-sub _get_perl_source_lines_by_real_path
-{
-    my ($real_path) = @_;
-    return _get_perl_source_lines_by_file_id( _get_perl_file_id_by_real_path( $real_path ) );
-}
 
 sub _get_real_path_by_normalized_perl_file_id
 {
@@ -1222,6 +1150,9 @@ sub _set_up_debugger
     $_script_charset = $set_up_data->{charset};
     _process_breakpoints_descriptors( $set_up_data->{breakpoints} );
 
+    $_enable_compile_time_breakpoints = 1 if $set_up_data->{enableCompileTimeBreakpoints};
+    $_enable_noninteractive_mode = 1 if $set_up_data->{enableNonInteractiveMode};
+
     my $start_mode = $set_up_data->{startMode};
 
     if ($start_mode eq 'RUN')
@@ -1264,54 +1195,66 @@ sub _process_breakpoints_descriptors
         $_loaded_breakpoints{$real_path}->{$line} = $descriptor;
         $_queued_breakpoints_files{$real_path} = 1;
     }
-    _apply_queued_breakpoints();
+    _set_break_points_for_files() if $ready_to_go;
 }
 
-#
-# Checks list of loaded and unset breakpoints
-#
-sub _apply_queued_breakpoints
+
+sub _set_break_points_for_files
 {
     return unless $ready_to_go;
-    foreach my $file_path (keys %_queued_breakpoints_files)
+    my ($paths_array) = @_;
+
+    $paths_array = [ keys %_queued_breakpoints_files ] unless $paths_array;
+    $paths_array = [ $paths_array ] unless ref $paths_array;
+    return unless @{$paths_array};
+
+    my $default_context = undef;
+
+    foreach my $real_path (@{$paths_array})
     {
-        _set_break_points_for_file( $file_path );
-    }
-}
+        _report "Setting breakpoints for %s", $real_path if $_dev_mode;
 
-sub _set_break_points_for_file
-{
-    my ($real_path) = @_;
+        # mapping real path to file id
+        my $perl_file_id = $real_path =~ /^\(eval \d+\)/
+            ? $real_path
+            : exists $_paths_to_perl_file_id_map{$real_path} ? $_paths_to_perl_file_id_map{$real_path} : next;
 
-    _report "Setting breakpoints for %s", $real_path if $_dev_mode;
-    my $loaded_breakpoints_descriptors = _get_loaded_breakpoints_by_real_path( $real_path ) or return;
-    my $perl_source_lines = _get_perl_source_lines_by_real_path( $real_path ) or return;
-    my $perl_breakpoints_map = _get_perl_line_breakpoints_map_by_real_path( $real_path ) or return;
-    my $perl_file_id = _get_perl_file_id_by_real_path( $real_path );
-    my $old_context = _switch_context( $perl_file_id );
+        # getting perl source lines and breakpoints
+        my $glob = $::{"_<$perl_file_id"};
+        next unless $glob && *{$glob}{ARRAY} && scalar @{*{$glob}{ARRAY}};
+        my $perl_source_lines = *{$glob}{ARRAY};
+        my $perl_breakpoints_map = *{$glob}{HASH};
 
-    #    print STDERR "Attempting to set breakpoints for $real_path\n";
-    my @lines = keys %$loaded_breakpoints_descriptors;
-    my $breakpoints_left = scalar @lines;
+        # getting breakpoints passed from the IDE
+        my $loaded_breakpoints_descriptors = _get_loaded_breakpoints_by_real_path( $real_path ) or next;
 
-    foreach my $real_line (@lines)
-    {
-        my $breakpoint_descriptor = $loaded_breakpoints_descriptors->{$real_line};
-        _report "Processing descriptor %s, %s, %s", @$breakpoint_descriptor{qw/path line remove/} if $_dev_mode;
+        # switching context
+        my $old_context = _switch_context( $perl_file_id );
+        $default_context //= $old_context;
 
-        if ($breakpoint_descriptor->{remove})
+        #    print STDERR "Attempting to set breakpoints for $real_path\n";
+        my @lines = keys %{$loaded_breakpoints_descriptors};
+        my $breakpoints_left = scalar @lines;
+
+        foreach my $real_line (@lines)
         {
-            $breakpoints_left -= _reset_breakpoint( $breakpoint_descriptor, $real_line, $perl_breakpoints_map );
-        }
-        else
-        {
-            $breakpoints_left -= _set_breakpoint( $breakpoint_descriptor, $real_line, $perl_breakpoints_map,
-                $perl_source_lines );
-        }
-    }
+            my $breakpoint_descriptor = $loaded_breakpoints_descriptors->{$real_line};
+            _report "Processing descriptor %s, %s, %s", @{$breakpoint_descriptor}{qw/path line remove/} if $_dev_mode;
 
-    delete $_queued_breakpoints_files{$real_path} unless $breakpoints_left;
-    _switch_context( $old_context );
+            if ($breakpoint_descriptor->{remove})
+            {
+                $breakpoints_left -= _reset_breakpoint( $breakpoint_descriptor, $real_line, $perl_breakpoints_map );
+            }
+            else
+            {
+                $breakpoints_left -= _set_breakpoint( $breakpoint_descriptor, $real_line, $perl_breakpoints_map,
+                    $perl_source_lines );
+            }
+        }
+
+        delete $_queued_breakpoints_files{$real_path} unless $breakpoints_left;
+    }
+    _switch_context( $default_context );
 }
 
 #sub mydie
@@ -1382,7 +1325,7 @@ sub step_handler
     $^W = 0;      # warnings are off
 
     # set breakpoints for evals if any appeared
-    _apply_queued_breakpoints() if $ready_to_go;
+    _set_break_points_for_files() if $ready_to_go;
 
     # updating current position
     my @caller = caller();
@@ -1519,7 +1462,7 @@ sub template_handler
         _report "Mapped template: %s to eval %s", $real_path, $eval_target if $_dev_mode;
 
         $_queued_breakpoints_files{$eval_target} = 1;
-        _apply_queued_breakpoints();
+        _set_break_points_for_files() if $ready_to_go;
     }
     else
     {
@@ -1538,7 +1481,7 @@ sub sub_handler
     {
         $_internal_process = 1;
 
-        _process_command( _get_next_command ) while _can_read;
+        _process_command( _get_next_command ) while $_enable_noninteractive_mode && _can_read;
         $old_db_single = $DB::single; # might be overriden in commands
 
         $DB::single = STEP_CONTINUE;
@@ -1696,8 +1639,7 @@ sub load_handler
         if $_debug_load_handler && $_dev_mode
     ;
 
-    _set_break_points_for_file( $real_path ) if $ready_to_go; # this is necessary, because perl internally re-initialize bp hash
-    _apply_queued_breakpoints() if $ready_to_go;
+    _set_break_points_for_files() if $ready_to_go;
 
     $_internal_process = $old_internal_process;
 
